@@ -3,7 +3,7 @@ use strict;
 use warnings;
 no warnings 'redefine';
 
-our $VERSION = do { my @r = ( q$Revision: 2.20 $ =~ /\d+/g ); sprintf "%d." . "%02d" x $#r, @r };
+our $VERSION = do { my @r = ( q$Revision: 2.22 $ =~ /\d+/g ); sprintf "%d." . "%02d" x $#r, @r };
 use Encode qw(find_encoding encode_utf8 decode_utf8);
 use MIME::Base64;
 use Carp;
@@ -44,40 +44,66 @@ my $re_encoded_text = qr/[^\?\s]*/;
 my $re_encoded_word = qr/=\?$re_charset(?:\*$re_language)?\?$re_encoding\?$re_encoded_text\?=/;
 my $re_capture_encoded_word = qr/=\?($re_charset)((?:\*$re_language)?)\?($re_encoding)\?($re_encoded_text)\?=/;
 
+our $STRICT_DECODE = 0;
+
 sub decode($$;$) {
     use utf8;
     my ( $obj, $str, $chk ) = @_;
 
-    # zap spaces between encoded words
-    1 while $str =~ s/($re_encoded_word)\s+($re_encoded_word)/$1$2/gos;
-
     # multi-line header to single line
     $str =~ s/(?:\r\n|[\r\n])([ \t])/$1/gos;
 
-    # concat consecutive encoded mime words with same charset, language and encoding
-    # fixes breaking inside multi-byte characters
-    1 while $str =~ s/$re_capture_encoded_word=\?\1\2\?\3\?($re_encoded_text)\?=/=\?$1$2\?$3\?$4$5\?=/o;
+    # decode each line separately
+    my @input = split /(\r\n|\r|\n)/o, $str;
+    my $output = substr($str, 0, 0); # to propagate taintedness
 
-    $str =~ s{$re_capture_encoded_word}{
-        if      (uc($3) eq 'B'){
-            $obj->{decode_b} or croak qq(MIME "B" unsupported);
-            decode_b($1, $4, $chk);
-        } elsif (uc($3) eq 'Q'){
-            $obj->{decode_q} or croak qq(MIME "Q" unsupported);
-            decode_q($1, $4, $chk);
-        } else {
-            croak qq(MIME "$3" encoding is nonexistent!);
-        }
-    }egox;
-    $_[1] = $str if $chk;
-    return $str;
+    while ( @input ) {
+
+        my $line = shift @input;
+        my $sep = shift @input;
+
+        # in strict mode encoded words must be always separated by spaces or tabs
+        # except in comments when separator between words and comment round brackets can be omitted
+        my $re_word_begin = $STRICT_DECODE ? qr/(?:[ \t\n]|\A)\(?/ : qr//;
+        my $re_word_sep = $STRICT_DECODE ? qr/[ \t]+/ : qr/\s*/;
+        my $re_word_end = $STRICT_DECODE ? qr/\)?(?:[ \t\n]|\z)/ : qr//;
+
+        # concat consecutive encoded mime words with same charset, language and encoding
+        # fixes breaking inside multi-byte characters
+        1 while $line =~ s/($re_word_begin)$re_capture_encoded_word$re_word_sep=\?\2\3\?\4\?($re_encoded_text)\?=(?=$re_word_end)/$1=\?$2$3\?$4\?$5$6\?=/;
+
+        $line =~ s{($re_word_begin)((?:$re_encoded_word$re_word_sep)*$re_encoded_word)(?=$re_word_end)}{
+            my $begin = $1;
+            my $words = $2;
+            $words =~ s{$re_capture_encoded_word$re_word_sep?}{
+                if (uc($3) eq 'B') {
+                    $obj->{decode_b} or croak qq(MIME "B" unsupported);
+                    decode_b($1, $4, $chk);
+                } elsif (uc($3) eq 'Q') {
+                    $obj->{decode_q} or croak qq(MIME "Q" unsupported);
+                    decode_q($1, $4, $chk);
+                } else {
+                    croak qq(MIME "$3" encoding is nonexistent!);
+                }
+            }eg;
+            $begin . $words;
+        }eg;
+
+        $output .= $line;
+        $output .= $sep if defined $sep;
+
+    }
+
+    $_[1] = '' if $chk; # empty the input string in the stack so perlio is ok
+    return $output;
 }
 
 sub decode_b {
-    my $enc  = shift;
-    my $d    = find_encoding($enc) or croak qq(Unknown encoding "$enc");
-    my $db64 = decode_base64(shift);
-    my $chk  = shift;
+    my ( $enc, $b, $chk ) = @_;
+    my $d = find_encoding($enc) or croak qq(Unknown encoding "$enc");
+    # MIME::Base64::decode_base64 ignores everything after a '=' padding character
+    # split string after each sequence of padding characters and decode each substring
+    my $db64 = join('', map { decode_base64($_) } split /(?<==)(?=[^=])/, $b);
     return $d->name eq 'utf8'
       ? Encode::decode_utf8($db64)
       : $d->decode( $db64, $chk || Encode::FB_PERLQQ );
@@ -95,25 +121,14 @@ sub decode_q {
 
 sub encode($$;$) {
     my ( $obj, $str, $chk ) = @_;
-
-    my @input = split /(\r\n|\r|\n)/o, $str;
-    my $output = substr($str, 0, 0); # to propagate taintedness
-
-    while ( @input ) {
-        my $line = shift @input;
-        my $sep = shift @input;
-        my $encoded = _encode_line($line, $obj);
-        $output .= _fold_line($encoded, $obj);
-        $output .= $sep if defined $sep;
-    }
-
-    return $output;
+    $_[1] = '' if $chk; # empty the input string in the stack so perlio is ok
+    return $obj->_fold_line($obj->_encode_line($str));
 }
 
 sub _fold_line {
-    my ($line, $obj) = @_;
+    my ( $obj, $line ) = @_;
     my $bpl = $obj->{bpl};
-    my $output = '';
+    my $output = substr($line, 0, 0); # to propagate taintedness
 
     while ( length $line ) {
         if ( $line =~ s/^(.{0,$bpl})(\s|\z)// ) {
@@ -132,57 +147,20 @@ sub _fold_line {
     return $output;
 }
 
-sub _encode_line {
-    my ($line, $obj) = @_;
-    my $bpl = $obj->{bpl};
-    my $output = '';
-
-    # try to detect header name and do not encode it
-    # needed for backward compatibility...
-    my $header_length = 0;
-    if ( $line =~ s/^([A-Za-z-]+:[ \t]+)//o ) {
-        $output .= $1;
-        $header_length = length($1);
-    }
-
-    my $safe_chars = q=!#$%&'*+,-./0-9:;A-Z[\\]^_`a-z{|}~=;
-    my $re_comment = qr/\([$safe_chars <>@]{1,$bpl}\)/;
-    my $re_addr = qr/[$safe_chars]{1,$bpl}\@[$safe_chars]{1,$bpl}/;
-    my $re_angle_addr = qr/<$re_addr>/;
-    my $re_not_encode = qr/(?:[ \t]+|\A)(?:$re_comment|$re_addr|$re_angle_addr)(?:[ \t]+|\z)/;
-
-    # try to detect substrings which looks like comment or email address and do not encode them if it is possible
-    # needed for backward compatibility but still should generate valid and correct mixed MIME encoded string
-    while ( $line =~ s/($re_not_encode)(.*)//o ) {
-        $output .= $obj->_encode($line, $header_length) if length $line;
-        $output .= $1;
-        $line = $2;
-        $header_length = 0 if $header_length;
-    }
-
-    $output .= $obj->_encode($line, $header_length) if length $line;
-
-    return $output;
-}
-
 use constant HEAD   => '=?UTF-8?';
 use constant TAIL   => '?=';
-use constant SINGLE => { B => \&_encode_b, Q => \&_encode_q, };
+use constant SINGLE => { B => \&_encode_b, Q => \&_encode_q, B_len => \&_encode_b_len, Q_len => \&_encode_q_len };
 
-sub _encode {
-    my ( $o, $str, $skip ) = @_;
+sub _encode_line {
+    my ( $o, $str ) = @_;
     my $enc  = $o->{encode};
+    my $enc_len = $enc . '_len';
     my $llen = ( $o->{bpl} - length(HEAD) - 2 - length(TAIL) );
 
-    # to coerce a floating-point arithmetics, the following contains
-    # .0 in numbers -- dankogai
-    my $skip_llen = ( $llen - $skip ) * ( $enc eq 'B' ? 3.0 / 4.0 : 1.0 / 3.0 );
-    $llen *= $enc eq 'B' ? 3.0 / 4.0 : 1.0 / 3.0;
     my @result = ();
     my $chunk  = '';
     while ( length( my $chr = substr( $str, 0, 1, '' ) ) ) {
-        use bytes ();
-        if ( bytes::length($chunk) + bytes::length($chr) > ( @result ? $llen : $skip_llen ) ) {
+        if ( SINGLE->{$enc_len}($chunk . $chr) > $llen ) {
             push @result, SINGLE->{$enc}($chunk);
             $chunk = '';
         }
@@ -196,17 +174,31 @@ sub _encode_b {
     HEAD . 'B?' . encode_base64( encode_utf8(shift), '' ) . TAIL;
 }
 
+sub _encode_b_len {
+    my ( $chunk ) = @_;
+    use bytes ();
+    return bytes::length($chunk) * 4 / 3;
+}
+
+my $valid_q_chars = '0-9A-Za-z !*+-/';
+
 sub _encode_q {
-    my $chunk = shift;
+    my ( $chunk ) = @_;
     $chunk = encode_utf8($chunk);
-    $chunk =~ s{
-	   ([^0-9A-Za-z])
-       }{
-            join("" => map {sprintf "=%02X", $_} unpack("C*", $1))
-       }egox;
+    $chunk =~ s{([^$valid_q_chars])}{
+        join("" => map {sprintf "=%02X", $_} unpack("C*", $1))
+    }egox;
+    $chunk =~ s/ /_/go;
     return HEAD . 'Q?' . $chunk . TAIL;
 }
 
+sub _encode_q_len {
+    my ( $chunk ) = @_;
+    use bytes ();
+    my $valid_count =()= $chunk =~ /[$valid_q_chars]/sgo;
+    return ( bytes::length($chunk) - $valid_count ) * 3 + $valid_count;
+}
+
 1;
 __END__
 
@@ -247,16 +239,24 @@ line.
 
 =head1 BUGS
 
-Before version 2.81 this module had broken both decoder and encoder.
-It caused that module inserted additional spaces and produced invalid
-MIME encoded strings. Decoder lot of times discarded white space
-characters, incorrectly interpreted data or decoded Base64 string as
-Quoted-Printable.
+Before version 2.83 this module had broken both decoder and encoder.
+Encoder inserted additional spaces, incorrectly encoded input data
+and produced invalid MIME strings. Decoder lot of times discarded
+white space characters, incorrectly interpreted data or decoded
+Base64 string as Quoted-Printable.
 
-As of version 2.81 encoder should be fully compliant of RFC 2047.
-Due to bugs in previous versions decoder is less strict and allows
-to decode also incorrectly encoded strings by previous versions of
-this module. But decoded strings will be slightly different.
+As of version 2.83 encoder should be fully compliant of RFC 2047.
+Due to bugs in previous versions of encoder, decoder is by default in
+less strict compatible mode. It should be able to decode strings
+encoded by pre 2.83 version of this module. But this default mode is
+not correct according to RFC 2047.
+
+In default mode decoder try to decode every substring which looks like
+MIME encoded data. So it means that MIME data does not need to be
+separated by white space. To enforce correct strict mode, set package
+variable $Encode::MIME::Header::STRICT_DECODE to 1, e.g. by localizing:
+
+C<require Encode::MIME::Header; local $Encode::MIME::Header::STRICT_DECODE = 1;>
 
 It would be nice to support encoding to non-UTF8, such as =?ISO-2022-JP?
 and =?ISO-8859-1?= but that makes the implementation too complicated.
